/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package Algoritmos;

import ALNS_Insercoes.InsercaoArrependimento;
import ALNS_Insercoes.InsercaoGuloso;
import ALNS_Insercoes.InsercaoHungaro;
import ALNS_Insercoes.InsercaoLambda;
import ALNS_Insercoes.InsercaoSemiGuloso;
import ALNS_Insercoes.Insercao_ILS;
import ALNS_Insercoes.Insercoes;
import ALNS_Remocoes.RemocaoAleatoria;
import ALNS_Remocoes.RemocaoMaiorCusto_Gulosa;
import ALNS_Remocoes.RemocaoMaiorCusto_SemiGulosa;
import ALNS_Remocoes.RemocaoMaqAleatoria;
import ALNS_Remocoes.RemocaoMaqMaior;
import ALNS_Remocoes.RemocaoShaw;
import ALNS_Remocoes.Remocoes;
import Heuristicas.CPG_ASPT;
import Metaheuristicas.MO_RVND;
import Principal.Dominancia;
import Principal.Instancia;
import Principal.Solucao;
import java.io.IOException;
import static java.lang.Double.min;
import java.util.ArrayList;
import java.util.Random;

/**
 *
 * @author luciano
 */

/*Algoritmo MO-ALNS*/
public class MO_ALNS {
    
    //
    private double t; //Tempo para execução
    private Instancia instancia; //Objeto contento os dados da instância do problema
    private Random rand;
    private int tamQ;
    private ArrayList<Solucao> D;
    
    //Métodos de inserção e remoção
    private ArrayList<Insercoes> insercoes;
    private ArrayList<Remocoes> remocoes;
    private static final double Q=0.05;
    
    //Atributor usados na aceitação do ALNS
    private double T; //temperatura
    private static final double alfa = 0.99; //fator para atualizar a temperatura
    
    //Learning Automata
    private ArrayList<Double> LA_insercoes;
    private ArrayList<Double> LA_remocoes;
    private ArrayList<Double> Prob_insercoes;
    private ArrayList<Double> Prob_remocoes;
    private static final double premio_Best = 0.2;
    private static final double premio_Current = 0.1;
    private static final double premio_Accepted = 0.05;
    private static final double penalidade = 0.02; 
    private int intervaloK;  
    private char estado; // B=best, C=current, A=accepted e P=not accepted 
    
    public MO_ALNS (double tempoExec, Instancia instancia, Random rand){
        
        
        this.t = tempoExec;
        this.instancia = instancia;
        this.rand = rand;
        this.D = new ArrayList<Solucao>();
        
        //Q***********************
        this.tamQ = Integer.max(4, (int) ((int) instancia.getNumJobs() * this.Q));
        System.out.println(this.tamQ);
       
        //Iniciando métodos de remoção
        this.remocoes = new ArrayList<Remocoes>();
        remocoes.add(new RemocaoAleatoria());
        remocoes.add(new RemocaoMaiorCusto_Gulosa());
        remocoes.add(new RemocaoMaiorCusto_SemiGulosa());
        remocoes.add(new RemocaoMaqMaior());
        remocoes.add(new RemocaoMaqAleatoria()); 
        remocoes.add(new RemocaoShaw());
        
        //Iniciando métodos de inserção
        this.insercoes = new ArrayList<Insercoes>();
        insercoes.add(new InsercaoLambda());
        insercoes.add(new InsercaoGuloso());
        insercoes.add(new InsercaoSemiGuloso());
        insercoes.add(new InsercaoHungaro());
        insercoes.add(new Insercao_ILS());
        insercoes.add(new InsercaoArrependimento()); 
                
        //Iniciando as Learning Automata
        LA_remocoes = new ArrayList<Double>();
        Prob_remocoes = new ArrayList<Double>();
        for(int i=0; i<remocoes.size(); i++){
            LA_remocoes.add(1.0);
            Prob_remocoes.add(100.0/remocoes.size());
        }    
        LA_insercoes = new ArrayList<Double>();
        Prob_insercoes = new ArrayList<Double>();
        for(int i=0; i<insercoes.size(); i++){
            LA_insercoes.add(1.0);
            Prob_insercoes.add(100.00/insercoes.size());
        }   
        
        //Intervalo K: número de heurísticas vezes 6
        this.intervaloK = Integer.max(remocoes.size(), insercoes.size()) * 6;
        
    }
    
    public void execute() throws IOException{
        
        
        //Iniciando contagem de tempo
        long tempoIni = System.currentTimeMillis(); // Iniciando Contagem do Tempo
        
        double tempoLimite = this.t * instancia.getNumJobs();
        
        //Criando soluções
        Solucao solAux = new Solucao(this.instancia);
        Solucao sol = new Solucao(this.instancia);
        Solucao sol2 = new Solucao(this.instancia);
        
        //Solução inicial
        sol = CPG_ASPT.execute(instancia, 0.2, rand); //Parcialmente guloso
        
        solAux.clonarSolucoes(sol);
        Dominancia.addSolucao(instancia, D, solAux);
        
        //Temperatura inicial: uma solução 10% pior que a solução inicial tem 50% de chance de ser escolhida
        double regraRL = (0.5 * sol.calcularMakespan()) + (0.5 * sol.calcularTEC());
        this.T = (0.05 * regraRL)/Math.log(2);
        
        //Enquanto houver tempo de execução
        long subTempo = 0; //Receberá o tempo gasto até o momento
        int k = 0; //controla a atualização da Learning Automata
        while(subTempo < (tempoLimite * 0.99) ){
            
            //
            sol2.clonarSolucoes(sol);
            
            System.out.print("Passei: ");
        
            //Método de Remoção
            int metodoRemocao = roleta(this.Prob_remocoes);
            System.out.print(metodoRemocao);
            ArrayList<Integer> conjQ = this.remocoes.get(metodoRemocao).execute(sol2, instancia, this.tamQ, rand);
            
            
            //Método de Inserção
            int metodoInsercao = roleta(this.Prob_insercoes);
            System.out.println(" " + metodoInsercao);
            sol2 = this.insercoes.get(metodoInsercao).execute(sol2, instancia, conjQ, rand);
            
            //Busca Local
            sol2 = MO_RVND.execute(this.instancia, sol2, rand, D);
            
            //Definindo dominância
            int dominancia = Dominancia.definirDominancia(sol2, sol);
            
            //Critério de aceitação
            regraRL = (0.5 * sol.calcularMakespan()) + (0.5 * sol.calcularTEC());
            double x = rand.nextDouble();
            boolean aceito = false;
            
            //Começa por estado não aceito!
            this.estado = 'P';
            
            //Se sol2 domina sol
            if(dominancia == 1 || dominancia == 2){
                this.estado = 'B';
                D = Dominancia.addSolucao(instancia, D, sol2);
            }
            //Se sol2 não domina sol e sol não domina sol2
            else if(dominancia == 5){
                this.estado = 'C';
                D = Dominancia.addSolucao(instancia, D, sol2);
            }
            //Se sol2 é aceito segundo o critério de aceitação
            else if(x < min(1,Math.exp(regraRL/T))){
                this.estado = 'A';
                aceito = true;
            }
            
            //Atualize rho+ e rho- na LA
            atualizaLA(metodoRemocao, metodoInsercao);
            
            //Seleciona aleatoriamente s em (D,s_duaslinhas)
            if(aceito){
                int r = rand.nextInt(D.size()+1);
                if(r == D.size())
                     sol.clonarSolucoes(sol2);
                else
                    sol.clonarSolucoes(D.get(r));
            }
            else{
                int r = rand.nextInt(D.size());
                sol.clonarSolucoes(D.get(r));
            }
                    
            //Atualiza periodicamente as probabilidades dos métodos da LA
            if(k == this.intervaloK){
                //Atualiza probabilidades
                atualizaProb();
                k = 0;
            }
            k++;
            
            //Atualiza a temperatura
            this.T = this.alfa * this.T;
            
            //Calcula o tempo gasto!
            long tempoFin = System.currentTimeMillis();
            subTempo = tempoFin - tempoIni;
            
        }
                
    }
    
    public void atualizaProb(){
        
        //Remoções
        double sum_Remocoes = 0;
        for(int i=0; i<LA_remocoes.size(); i++)
            sum_Remocoes += LA_remocoes.get(i);
        for(int i=0; i<Prob_remocoes.size(); i++)
            Prob_remocoes.set(i, (LA_remocoes.get(i) / sum_Remocoes) * 100);
        
        //Inserções
        double sum_Inserções = 0;
        for(int i=0; i<LA_insercoes.size(); i++)
            sum_Inserções += LA_insercoes.get(i);
        for(int i=0; i<Prob_insercoes.size(); i++)
            Prob_insercoes.set(i, (LA_insercoes.get(i) / sum_Inserções) * 100);
    }
    
    public void atualizaLA(int metodoRemocao, int metodoInsercao){
        
        if(this.estado == 'P')
            this.aplicarPenalidade(metodoRemocao, metodoInsercao);
        else
            this.aplicarPremio(metodoRemocao, metodoInsercao);
    }
    
    public void aplicarPremio(int metodoRemocao, int metodoInsercao){
    
        //Prêmio
        double premio = 0;
        if(estado == 'B')
            premio = this.premio_Best;
        else if(estado == 'C')
            premio = this.premio_Current;
        else if(estado == 'A')
            premio = this.premio_Accepted;
        
        //Learning Automata - Remoções
        for(int i=0; i<LA_remocoes.size(); i++){
            if(metodoRemocao == i){
                double auxValue = LA_remocoes.get(i) + premio * (1 - LA_remocoes.get(i));
                LA_remocoes.set(i, auxValue);
            }
            else{
                double auxValue = (1 - premio) * LA_remocoes.get(i);
                LA_remocoes.set(i, auxValue);
            }
        }
        
        //Learning Automata - Inserções
        for(int i=0; i<LA_insercoes.size(); i++){
            if(metodoInsercao == i){
                double auxValue = LA_insercoes.get(i) + premio * (1 - LA_insercoes.get(i));
                LA_insercoes.set(i, auxValue);
            }
            else{
                double auxValue = (1 - premio) * LA_insercoes.get(i);
                LA_insercoes.set(i, auxValue);
            }
        }
    }
    
    public void aplicarPenalidade(int metodoRemocao, int metodoInsercao){
    
        //Learning Automata - Remoções
        for(int i=0; i<LA_remocoes.size(); i++){
            if(metodoRemocao == i){
                double auxValue = (1 - this.penalidade) * LA_remocoes.get(i);
                LA_remocoes.set(i, auxValue);
            }
            else{
                double auxValue = (this.penalidade / (this.LA_remocoes.size() - 1)) + (1 - this.penalidade) * LA_remocoes.get(i);
                LA_remocoes.set(i, auxValue);
            }
        }
        
        //Learning Automata - Inserções
        for(int i=0; i<LA_insercoes.size(); i++){
            if(metodoInsercao == i){
                double auxValue = (1 - this.penalidade) * LA_insercoes.get(i);
                LA_insercoes.set(i, auxValue);
            }
            else{
                double auxValue = (this.penalidade / (this.LA_insercoes.size() - 1)) + (1 - this.penalidade) * LA_insercoes.get(i);
                LA_insercoes.set(i, auxValue);
            }
        }
    }
    
    public int roleta(ArrayList<Double> conj_Prob){
        
        //Roda a roleta!
        double randRoleta = 100 * (this.rand.nextDouble());
        
        double acumulado = 0;
        int metodoEscolhido = 0;
        for(int i=0; i<conj_Prob.size(); i++){
            acumulado += conj_Prob.get(i);
            if(acumulado > randRoleta){
                metodoEscolhido = i;
                break;
            }    
        }
        
        return metodoEscolhido;
    }

    public ArrayList<Solucao> getD() {
        return D;
    }

    public void setD(ArrayList<Solucao> D) {
        this.D = D;
    }
    
    
    
    
}
