/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package Algoritmos;

import Partes_MOALNSD.LA_ALNS_MULT;
import Partes_MOALNSD.CPG_ASPT_MULT;
import Partes_MOALNSD.Pesos;
import Principal.Solucao;
import java.util.ArrayList;
import java.util.Collections;
import Principal.Dominancia;
import Principal.Instancia;
import java.io.IOException;
import java.util.Objects;
import java.util.Random;

import Partes_MOALNSD.Scheffe;
import Partes_MOALNSD.Scheffe;

/**
 *
 * @author luciano
 */

/*Algoritmo MO-ALNS/D*/
public class MO_ALNS_D {
    
    private Instancia instancia;
    private Random rand;
    
    private int n; //Total de Subproblemas
    private static final int T=2; //Total de vizinhos
    
    private ArrayList<Solucao> pop; //População    
    private ArrayList<ArrayList<Integer>> B; //Vetor de vizinhos de cada Subproblema
    
    private ArrayList<Pesos> vetorPesos;
    
    public MO_ALNS_D (int n, Instancia instancia, Random rand){
        this.n = n;
        this.B = new ArrayList<ArrayList<Integer>>();
        this.pop = new ArrayList<Solucao>();
        this.instancia = instancia;
        this.rand = rand;
    }
    
    public int getN() {
        return n;
    }

    public void setN(int n) {
        this.n = n;
    }

    public ArrayList<Solucao> getPop() {
        return pop;
    }

    public void setPop(ArrayList<Solucao> pop) {
        this.pop = pop;
    }

    public ArrayList<ArrayList<Integer>> getB() {
        return B;
    }

    public void setB(ArrayList<ArrayList<Integer>> B) {
        this.B = B;
    }

    public ArrayList<Pesos> getVetorPesos() {
        return vetorPesos;
    }

    public void setVetorPesos(ArrayList<Pesos> vetorPesos) {
        this.vetorPesos = vetorPesos;
    }
    
    
    public void execute(int geracoes, int t_ALNS) throws IOException{
        
        double tempoLimite_ALNS = (instancia.getNumJobs() * t_ALNS);
        
        //Gerar vetor de pesos
        ArrayList<Pesos> vetor = Scheffe.executar(this.getN());
        this.setVetorPesos(vetor);
                
        //Calcular vizinhos para cada subproblema
        this.gerarVizinhos(vetorPesos);
        //vetorPesos = Collections.shuffle(vetorPesos);
        
        //Gerar a população inicial
        for(int i=0; i<this.getN(); i++){
            pop.add(CPG_ASPT_MULT.execute(this.instancia, this.vetorPesos.get(i).getPeso_Cmax(), vetorPesos.get(i).getPeso_TEC(), 0.2, new Random()));
        }
        
        //Armazena o z* na execução do MOEAD
        double z_Cmax_MOEAD = Double.MAX_VALUE;
        double z_TEC_MOEAD  = Double.MAX_VALUE;
        
        //Calcular o z* para cada objetivo
        for(int i=0; i<this.getN(); i++){
            
            //Atualizando o z* para o Cmax
            if(z_Cmax_MOEAD > pop.get(i).calcularMakespan())
                z_Cmax_MOEAD = pop.get(i).calcularMakespan();
            if(z_TEC_MOEAD > pop.get(i).getTEC())
                z_TEC_MOEAD = pop.get(i).getTEC();
        
        }
        
        //Atualiza o z* para todas soluções da População e Atualizar o max para todas soluções da População
        for(int i=0; i<this.getN(); i++){
            
            /** Z* **/
            this.pop.get(i).setZ_Cmax(z_Cmax_MOEAD);
            this.pop.get(i).setZ_Tec(z_TEC_MOEAD);
        }
        
        LA_ALNS_MULT alns                = new LA_ALNS_MULT(tempoLimite_ALNS, instancia, rand);
        ArrayList<Double> LA_remocoes  = alns.getLA_remocoes();
        ArrayList<Double> LA_insercoes = alns.getLA_insercoes();
        
        //Enquanto houver tempo de execução
        long iter_Geracoes = 0; //Receberá o tempo gasto até o momento
        while(iter_Geracoes < geracoes ){
           
            iter_Geracoes++;
            
            for(int i=0; i<n; i++){
               
                //Seleciona um vizinho aleatório do subproblema i
                int r = rand.nextInt(T); //Seleciona um índice para o vizinho (T = total de vizinhos)
                int v = this.getB().get(i).get(r); //Pega um vizinho aleatório
                
                alns.setLA_remocoes(LA_remocoes);
                alns.setLA_insercoes(LA_insercoes);
                alns.atualizaProb();
                //Executa
                Solucao v_Linha = alns.execute(pop.get(v));
                //Armazena as LAs atuais
                LA_remocoes    = alns.getLA_remocoes();
                LA_insercoes   = alns.getLA_insercoes();
                
                //Atualiza Z
                boolean atualizou = false;
                if(z_Cmax_MOEAD > v_Linha.calcularMakespan()){
                    z_Cmax_MOEAD = v_Linha.calcularMakespan();
                    atualizou = true;
                }   
                
                //ALT
                if(z_TEC_MOEAD > v_Linha.getTEC()){
                    z_TEC_MOEAD = v_Linha.getTEC();
                    atualizou = true;
                }
                
                if(atualizou){
                    //Atualiza o z* para todas soluções da População
                    for(int j=0; j<this.getN(); j++){
                        this.pop.get(j).setZ_Cmax(z_Cmax_MOEAD);
                        this.pop.get(j).setZ_Tec(z_TEC_MOEAD);
                    }
                }
                
                int j = i; 
                                                
                //Avaliação
                for(int g=0; g<this.getB().get(j).size(); g++){
                    
                    //Verique se v' é melhor que algum dos vizinhos do subproblema i
                    int vizinho = this.getB().get(j).get(g); //Vizinho de i
                    
                    v_Linha.setPeso_Cmax(pop.get(vizinho).getPeso_Cmax());
                    v_Linha.setPeso_TEC(pop.get(vizinho).getPeso_TEC());
                    
                    if(pop.get(vizinho).calcularFO2() > v_Linha.calcularFO2()){

                        Solucao vizNew = new Solucao(instancia);
                        vizNew.clonarSolucoes(v_Linha);
                        pop.set(vizinho, vizNew);
                        
                    }
                }
                
            }
            
        }
        
        
        
    }
    
    public void gerarVizinhos(ArrayList<Pesos> vetorPesos){
        
        //
        ArrayList<ArrayList<Double>> distEuclidiana = new ArrayList<ArrayList<Double>>();        
        for(int p=0; p<this.getN(); p++){
            
            //Peso do Ponto p
            double pMakespan = vetorPesos.get(p).getPeso_Cmax();
            double pSumPond  = vetorPesos.get(p).getPeso_TEC();
            distEuclidiana.add(new ArrayList<Double>());
                        
            for(int q=0; q<this.getN(); q++){
                
                //Peso do Ponto q
                double qMakespan = vetorPesos.get(q).getPeso_Cmax();
                double qSumPond  = vetorPesos.get(q).getPeso_TEC();
            
                //Cálculo da distância Euclidiana de "p" para "q"
                double dist = Math.sqrt( Math.pow((pMakespan - qMakespan), 2) + Math.pow((pSumPond - qSumPond), 2) );   
                distEuclidiana.get(p).add(dist);
            }
        }
        
        for(int i=0; i<this.getN(); i++){
            
            double vizinho1 = Double.MAX_VALUE;
            int indVizinho1 = Integer.MAX_VALUE;
            
            double vizinho2 = Double.MAX_VALUE;
            int indVizinho2 = Integer.MAX_VALUE;
            
            //Vizinho 1
            for(int z=0; z<distEuclidiana.get(i).size(); z++){
                if(z != i){
                    //Pega o vizinho mais próximo
                    if(vizinho1 > distEuclidiana.get(i).get(z)){
                        vizinho1    = distEuclidiana.get(i).get(z);
                        indVizinho1 = z;
                    }
                }
            }
            
            //Vizinho 2
            for(int z=0; z<distEuclidiana.get(i).size(); z++){
                if(z != i && z != indVizinho1){
                    //Pega o segundo vizinho mais próximo
                    if(vizinho2 > distEuclidiana.get(i).get(z)){
                        vizinho2    = distEuclidiana.get(i).get(z);
                        indVizinho2 = z;
                    }
                }
            }
            
            //Adiciona na lista de vizinhos
            this.B.add(new ArrayList<Integer>());
            this.B.get(i).add(indVizinho1);
            this.B.get(i).add(indVizinho2);
            
        }
        
        
    }
    
    //Modelo do gráfico Pareto
    public void imprimirPop(){
        
        ArrayList<Double> obj1 = new ArrayList<Double>();
        ArrayList<Double> obj2 = new ArrayList<Double>();
        
        boolean []valida = new boolean[pop.size()];
        for(int w=0; w<pop.size(); w++){
            valida[w] = true;
        }
        
        System.out.println("Makespan;TEC");
        for(int i=0; i<pop.size(); i++){
            
            double menorMakespan = Double.MAX_VALUE;
            int indice = -1;
            
            for(int j=0; j<pop.size(); j++){
                if(pop.get(j).calcularMakespan() < menorMakespan && valida[j]){
                    menorMakespan = pop.get(j).calcularMakespan();
                    indice = j;
                }
            }
            
            valida[indice] = false;            
            
            obj1.add(pop.get(indice).calcularMakespan());
            obj2.add(pop.get(indice).calcularTEC());
        }
        
    }
    
    
    
    
}
