/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package Algoritmos;

import ALNS_Remocoes.RemocaoMaqMaior;
import Heuristicas.CPG_ASPT;
import Principal.Dominancia;
import Principal.Instancia;
import Principal.Solucao;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Random;

import ALNS_Insercoes.InsercaoArrependimento;
import ALNS_Insercoes.InsercaoGuloso;
import ALNS_Insercoes.InsercaoHungaro;
import ALNS_Insercoes.InsercaoLambda;
import ALNS_Insercoes.InsercaoSemiGuloso;
import ALNS_Insercoes.Insercao_ILS;
import ALNS_Insercoes.Insercoes;
import ALNS_Remocoes.RemocaoAleatoria;
import ALNS_Remocoes.RemocaoMaiorCusto_Gulosa;
import ALNS_Remocoes.RemocaoMaiorCusto_SemiGulosa;
import ALNS_Remocoes.RemocaoMaqAleatoria;
import ALNS_Remocoes.RemocaoMaqMaior;
import ALNS_Remocoes.RemocaoShaw;
import ALNS_Remocoes.Remocoes;
import Metaheuristicas.MO_RVND;
import static java.lang.Double.min;

/**
 *
 * @author luciano
 */

/*Algoritmo MO-LNS*/
public class MO_LNS {
    
    private double t; //Tempo para execução
    private Instancia instancia; //Objeto contento os dados da instância do problema
    private Random rand;
    private int tamQ;
    private ArrayList<Solucao> D;
    private double T; //temperatura
    private static final double alfa = 0.99; //fator para atualizar a temperatura
        
    public MO_LNS (double tempoExec, Instancia instancia, Random rand){
        this.t = tempoExec;
        this.instancia = instancia;
        this.rand = rand;
        this.D = new ArrayList<Solucao>();
    }
    
    public void execute() throws IOException{
        
        //Iniciando contagem de tempo
        long tempoIni = System.currentTimeMillis(); // Iniciando Contagem do Tempo
        
        //Tempo limite: NOVO
        double tempoLimite = this.t * instancia.getNumJobs();
        
        //Criando soluções
        Solucao solAux = new Solucao(this.instancia);
        Solucao sol = new Solucao(this.instancia);
        Solucao sol2 = new Solucao(this.instancia);
        
        //Solução inicial
        sol = CPG_ASPT.execute(instancia, 0.2, rand); //Parcialmente guloso
        
        //Adicionando solução à D
        solAux.clonarSolucoes(sol);
        Dominancia.addSolucao(instancia, D, solAux);
        
        //Temperatura inicial: uma solução 10% pior que a solução inicial tem 50% de chance de ser escolhida
        double regraRL = (0.5 * sol.calcularMakespan()) + (0.5 * sol.calcularTEC());
        this.T = (0.05 * regraRL)/Math.log(2);
        
        //Enquanto houver tempo de execução
        long subTempo = 0; //Receberá o tempo gasto até o momento
        while(subTempo < (tempoLimite * 0.99) ){
            
            //
            sol2.clonarSolucoes(sol);
            
            //Remove tarefas => Método Remoção Máquina Aleatória
            Remocoes remocao = new RemocaoMaqAleatoria();            
            ArrayList<Integer> conjQ = remocao.execute(sol2, instancia, this.tamQ, rand);
            
            //Insere tarefas => Inserção Gulosa
            Insercoes insercao = new InsercaoGuloso();
            sol2 = insercao.execute(sol2, instancia, conjQ, rand);
            
            //Busca Local
            sol2 = MO_RVND.execute(this.instancia, sol2, rand, D);
            
            //Definindo dominância
            int dominancia = Dominancia.definirDominancia(sol2, sol);
            
            //Critério de aceitação
            regraRL = (0.5 * sol.calcularMakespan()) + (0.5 * sol.calcularTEC());
            double x = rand.nextDouble();
            boolean aceito = false;
            
            //Se sol2 domina sol
            if(dominancia == 1 || dominancia == 2){
                D = Dominancia.addSolucao(instancia, D, sol2);
            }
            //Se sol2 não domina sol e sol não domina sol2
            else if(dominancia == 5){
                D = Dominancia.addSolucao(instancia, D, sol2);
            }
            //Se sol2 é aceito segundo o critério de aceitação
            else if(x < min(1,Math.exp(regraRL/T))){
                aceito = true;
            }
            
            //Seleciona aleatoriamente s em (D,s_duaslinhas)
            if(aceito){
                int r = rand.nextInt(D.size()+1);
                if(r == D.size())
                     sol.clonarSolucoes(sol2);
                else
                    sol.clonarSolucoes(D.get(r));
            }
            else{
                int r = rand.nextInt(D.size());
                sol.clonarSolucoes(D.get(r));
            }
            
            //Atualiza a temperatura
            this.T = this.alfa * this.T;
            
            //Calcula o tempo gasto!
            long tempoFin = System.currentTimeMillis();
            subTempo = tempoFin - tempoIni;
            
        }
        
    }

    public ArrayList<Solucao> getD() {
        return this.D;
    }
    
}
