/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Heuristicas;

import Principal.Instancia;
import Principal.Solucao;
import java.util.ArrayList;
import java.util.Random;

/**
 *
 * @author DecomLuciano
 *
 * Contrutivo Guloso ASPT (Adaptive Shortest Processing Time) */
public class CG_ASPT {
    
    public static Solucao execute (Instancia instancia){
        
        //Cria a Solução
        Solucao sol = new Solucao(instancia);
        
        //Mudança (05/10/2017): Inicia todas as tarefas com índice de velocidade aleatório
        for(int j=0;j<instancia.getNumJobs();j++){
            sol.alterarSpeed(j, 0);        
        }
       
        //Conjunto de Tarefas Não Alocada
        ArrayList<Integer> conjTarNaoAloc = new ArrayList<Integer>();
        for(int i=0; i < instancia.getNumJobs(); i++){
            conjTarNaoAloc.add(i);
        }
        
        //Enquanto existirem tarefas não alocadas
        while(conjTarNaoAloc.size() > 0){
            
            //Definindo o custo Acumulado
            double [][]custoAc = new double[instancia.getNumJobs()+1][instancia.getNumMaqs()+1];
            for(int i=0; i<conjTarNaoAloc.size();i++){
                for(int j=0; j<instancia.getNumMaqs();j++){
                    int job = conjTarNaoAloc.get(i); //Pegando uma Tarefa
                    custoAc[job][j] = CG_ASPT.defineCustoAcu(instancia, job, j, sol); //Quanto menor a pontuação, menor é o tempo acumulado com a inserção da tarefa
                }
            }
            
            //Pegando o menor custo
            int maqEscolhida = -1;
            int tarEscolhida = -1;
            double menorCusto = 99999999;
            for(int i=0; i<conjTarNaoAloc.size();i++){
                for(int j=0; j<instancia.getNumMaqs();j++){
                    int job = conjTarNaoAloc.get(i); //Pegando uma Tarefa
                    //Menor custo
                    if(custoAc[job][j] < menorCusto){
                        maqEscolhida = j;
                        tarEscolhida = job;
                        menorCusto = custoAc[job][j];
                    }
                }
            }
            
            //Adicionar Tarefa a Solução
            sol.getSol().get(maqEscolhida).add(tarEscolhida);
            
            //Retirar Tarefa do conjunto de tarefas não alocadas
            for(int i=0; i<conjTarNaoAloc.size();i++){
                int job = conjTarNaoAloc.get(i); //Pegando uma Tarefa
                if(job == tarEscolhida){
                    conjTarNaoAloc.remove(i);
                }
            }
        }
        
        //Define o custo para as máquinas
        sol.definirCustoMaqs();
        
        return sol;
        
    }
    
    public static double defineCustoAcu(Instancia instancia, int job, int machine, Solucao solucao){
        
        //Custo Inicial
        double custo = solucao.calcularCustoMaq(machine);
        
        //Verificando se a máquina ainda esta vazia
        if(solucao.getSol().get(machine).isEmpty()){
            custo += instancia.getPrep()[machine][job][job] + instancia.getProc()[machine][job];
        }
        //A Máquina já tem itens alocados
        else{
            int ultimaTarMaq = solucao.getSol().get(machine).get(solucao.getSol().get(machine).size()-1);
            custo += (double) instancia.getPrep()[machine][ultimaTarMaq][job] + instancia.getProc()[machine][job];
        }
        
        return custo;        
    }
    
}
