/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package Principal;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;

/**
 *
 * @author DecomLuciano
 */

//Classe responsável por armazenar uma Solução e processá-la
public class Solucao implements Comparable<Solucao>, Serializable{
    
    private ArrayList<ArrayList<Integer>>  sol; //Estrutura de armazenamento da Solução
    
    private double makespan; //Makespan da Solução
    private ArrayList<Double> custoMaq; //Contém o custo de todas as máquinas
    private Instancia instancia; //Instância Atual
    
    //Novas características
    private double TEC; //Consumo total de energia
    //Estrutura de aramazenamento das Velocidades
    private int []speed; //Cada posição representa uma tarefa
    
    //Novos: Dezembro de 2017 -> MOEA/D
    private double peso_Cmax;
    private double peso_TEC;
    private double z_Cmax;
    private double z_Tec;
    
    public Solucao(Instancia instancia){
        
        this.instancia = instancia;
        this.makespan = 999999;

        this.sol = new ArrayList<ArrayList<Integer>>();
        this.custoMaq = new ArrayList<Double>();
        for(int i=0;i<instancia.getNumMaqs();i++){
            ArrayList<Integer> aux = new ArrayList<Integer>();
            this.sol.add(i,aux);
            this.custoMaq.add(i, 0.0);
        }
        
        //Novas características
        this.TEC = Double.MAX_VALUE;
        this.speed = new int[instancia.getNumJobs()];
        
    }
    
    public Solucao(){
        this.instancia = instancia;
        this.makespan = 999999;
        this.sol = new ArrayList<ArrayList<Integer>>();
        this.custoMaq = new ArrayList<Double>();
        
        //Novas características
        this.TEC = Double.MAX_VALUE;
        this.speed = new int[0];
    }
    
    public double calcularFO2(){
        
        //Makespan
        double cmax_FO = this.calcularMakespan();
        double TEC_FO  = this.getTEC();
        
        if(cmax_FO < this.getZ_Cmax())
            this.setZ_Cmax(cmax_FO);
        if(TEC_FO < this.getZ_Tec())
            this.setZ_Tec(TEC_FO);
       
        
        /**** FUNÇÕES DE DECOMPOSIÇÃO DO MOEA/D ****/
        
        //Tchebycheff Approach
        double valorFuncDecomposicao = Math.max((this.getPeso_Cmax() * (cmax_FO - this.getZ_Cmax())), (this.getPeso_TEC() * (TEC_FO - this.getZ_Tec())));
    
        //System.out.println(valorFuncDecomposicao);
        
        return valorFuncDecomposicao;
    
    }
    
    
    public double getPeso_Cmax() {
        return peso_Cmax;
    }

    public void setPeso_Cmax(double peso_Cmax) {
        this.peso_Cmax = peso_Cmax;
    }

    public double getPeso_TEC() {
        return peso_TEC;
    }

    public void setPeso_TEC(double peso_TEC) {
        this.peso_TEC = peso_TEC;
    }

    public double getZ_Cmax() {
        return z_Cmax;
    }

    public void setZ_Cmax(double z_Cmax) {
        this.z_Cmax = z_Cmax;
    }

    public double getZ_Tec() {
        return z_Tec;
    }

    public void setZ_Tec(double z_Tec) {
        this.z_Tec = z_Tec;
    }
    
    public void alterarSpeed(int tarefa, int newSpeed){
        this.speed[tarefa] = newSpeed;
    }
    
    public double calcularTEC(){
        
        double sum = 0;
        
        //Percorrendo todas as máquinas
        for(int i=0; i<sol.size();i++){
            for(int j=0; j<sol.get(i).size(); j++){
                
                //Tempo de processamento dividido pelo índice de velocidade
                int tarefa = this.sol.get(i).get(j);
                double custoProcessamento = this.instancia.getProc()[i][tarefa]/instancia.getV()[this.speed[tarefa]];
                
                //Potência dividido por 60 
                double pi = (double) instancia.getPi()[i]/60;
                
                //Fórmula: lambda_{l} * pi_{i}/60 * p_{ik}/v_{l} 
                sum += instancia.getLambda()[this.speed[tarefa]] * pi * custoProcessamento; 
              
            
            }
            
        }
         
        //Converte em Float para desprezar um número muito grande de casas decimais
        this.setTEC((float)sum);
        
        return this.TEC;
    
       
        
    }

    public int[] getSpeed() {
        return speed;
    }

    public void setSpeed(int[] speed) {
        this.speed = speed;
    }

    public double getTEC() {
        return TEC;
    }

    public void setTEC(double TEC) {
        this.TEC = TEC;
    }
    
    
    
    public void setMakespan(double valor){
        this.makespan = valor;
    }

    public ArrayList<ArrayList<Integer>> getSol() {
        return sol;
    }

    public void setSol(ArrayList<ArrayList<Integer>> solucao) {
        this.sol = solucao;
    }
    
    public void definirCustoMaqs(){
        
        //Parte antiga
        ArrayList<Double> custoMaquina = new ArrayList<Double>();
        for(int i=0;i<this.sol.size();i++){
            custoMaquina.add(i, this.calcularCustoMaq(i)); 
        }
        this.custoMaq = custoMaquina; //Atualiza o custo de todas as máquinas
        
        
    }
    
    public void imprimirSolucao(){
        System.out.println("Imprimindo Solução");
        for(int i=0; i<sol.size();i++){
            System.out.print("M" + i + ":");
            for(int j=0; j<sol.get(i).size(); j++){
                if(j == 0)
                    System.out.print(this.sol.get(i).get(j));
                else
                    System.out.print(" | " + this.sol.get(i).get(j));
            }
            System.out.print(" Custo:" + this.getCustoMaq().get(i) + " CustoPI:" + this.calcularCustoMaq(i));
            
            double valor_1 = this.getCustoMaq().get(i);
            int valor_1_int = (int) valor_1;
            double valor_2 = this.calcularCustoMaq(i);
            int valor_2_int = (int) valor_2;
            
            //if(this.getCustoMaq().get(i) != this.calcularCustoMaq(i))
            if(valor_1_int != valor_2_int)
                System.out.println("ERRO Valor Inválido!!!");
            System.out.println();
        }
        int total = 0;
        for(int i2=0; i2<this.getSol().size(); i2++){
            total += this.getSol().get(i2).size();
        }
        if(total != instancia.getNumJobs())
            System.out.println("ERRO => Total de Tarefas:" + total);
        
        //Imprimir vetor de índice de velocidades
        System.out.print("Índice de Velocidades: ");
        for(int j=0; j<instancia.getNumJobs(); j++){
            System.out.print(this.speed[j] + " ");
        }
        System.out.println();
        
        //calcularMakespan();
        //System.out.println("Makespan: " + this.makespan);
        //System.out.println("TEC: " + this.TEC);
        System.out.println("Makespan: " + this.calcularMakespan());
        System.out.println("TEC: " + this.calcularTEC() + " == " + this.TEC);
    
    }
    
    //Faz o papel do getMakespan()
    public double calcularMakespan(){
        //Converte em Float para desprezar um número muito grande de casas decimais
        this.makespan = new Float (Collections.max(this.custoMaq)); //Atualiza o makespan
        return this.makespan;
    }
    
    //Regra delta para velocidade: Calcula o custo de troca de uma velocidade
    public double regraDelta_Velocidade(Instancia instancia, int maq, int tar, int novaVel){
        
        
        //Na velocidade antiga
        double custoAntigo = (instancia.getProc()[maq][tar]/instancia.getV()[this.speed[tar]]);
        
        //Na velocidade nova
        double custoNovo = (instancia.getProc()[maq][tar]/instancia.getV()[novaVel]);
        
      
        
        //Retorna o novo custo em unidades de tempo
        return custoNovo - custoAntigo;
    }
    
    
    //Regra Delta Inserção
    public double regraDelta_Inserir2(Instancia instancia, int maq, int tar, int posTar){
   
        //Tarefa Anterior
        int tarAnt = -1;
        if(posTar > 0)
            tarAnt = this.getSol().get(maq).get(posTar-1);
        
        //Tarefa Posterior
        int tarDep = -1; //Tarefa Sucessora
        if(posTar < this.getSol().get(maq).size()-1)
            tarDep = this.getSol().get(maq).get(posTar+1);
        
        //Obs: -1 indica que não existe a posição
        
        double custo = 0;
        
        if(tarAnt != -1){ //Tarefa Anterior
            custo += instancia.getPrep()[maq][tarAnt][tar];
        }
        //Se for o primeiro tem de somar o tempo de preparação
        else if(tarAnt == -1){
            custo += instancia.getPrep()[maq][tar][tar];
        }
                
        //Custo de Processamento
        double custoProcessamento = (instancia.getProc()[maq][tar]/instancia.getV()[this.speed[tar]]); //Tarefa Inserida
        custo += custoProcessamento;
        
               
        if(tarDep != -1){
            custo += instancia.getPrep()[maq][tar][tarDep]; //Tarefa Sucessora
        }
        
        if(tarAnt != -1 && tarDep != -1){
            custo -= instancia.getPrep()[maq][tarAnt][tarDep]; //Prep Tarefa Anterior para Tarefa Sucessora
        }
        else if(tarAnt == -1  && tarDep != -1){
            custo -= instancia.getPrep()[maq][tarDep][tarDep]; //Tarefa Sucessora é a 1ª
        }
        
        return custo;
    }
    
    //Regra Delta Retirada
    public double regraDelta_Retirar2(Instancia instancia, int maq, int tar, int posTar){
        
        //Tarefa Anterior
        int tarAnt = -1;
        if(posTar > 0)
            tarAnt = this.getSol().get(maq).get(posTar-1);
        
        //Tarefa Posterior
        int tarDep = -1; //Tarefa Sucessora
        if(posTar < this.getSol().get(maq).size()-1)
            tarDep = this.getSol().get(maq).get(posTar+1); 
   
        //Obs: -1 indica que não existe a posição
        
        double custo = 0;
        
        if(tarAnt != -1){ //Tarefa Anterior
            custo -= instancia.getPrep()[maq][tarAnt][tar];
        }
        //Se for o primeiro tem de subtrair o tempo de preparação
        else if(tarAnt == -1){
            custo -= instancia.getPrep()[maq][tar][tar];
        }
        
        //Custo de Processamento!
        double custoProcessamento = (instancia.getProc()[maq][tar]/instancia.getV()[this.speed[tar]]); //Tarefa Removida
        custo -= custoProcessamento;
        
        //
        //Fórmula: lambda_{l} * pi_{i}/60 * p_{ik}/v_{l} 
        //double custoRemocao_TEC = (instancia.getLambda()[this.speed[tar]]) * ((double) instancia.getPi()[maq]/60) * (custoProcessamento); 
        //double custoAntigo_TEC = this.getTEC();
        //this.setTEC(custoAntigo_TEC - custoRemocao_TEC);
        
        
        //Original: Dezembro de 2017!
        //custo -= (instancia.getProc()[maq][tar]/instancia.getV()[this.speed[tar]]); //Tarefa Removida
        
        if(tarDep != -1){
            custo -= instancia.getPrep()[maq][tar][tarDep]; //Tarefa Sucessora
        }
        
        if(tarAnt != -1 && tarDep != -1){
            custo += instancia.getPrep()[maq][tarAnt][tarDep]; //Prep Tarefa Anterior para Tarefa Sucessora
        }
        else if(tarAnt == -1  && tarDep != -1){
            custo += instancia.getPrep()[maq][tarDep][tarDep]; //Tarefa Sucessora é a 1ª
        }
        
        return custo;
    }
    
   

    //Modificado 02/10/2017
    //Calcula o custo de uma máquina, usando a solução atual 
    public double calcularCustoMaq(int numMaq){
    
        double custoAux = 0;
        int ultimaTarefa = 0;
        for(int j=0; j<sol.get(numMaq).size(); j++){
            int tarefa = sol.get(numMaq).get(j);
            if(j == 0){
                //System.out.println(" j==0 " + this.instancia.getPrep()[numMaq][tarefa][tarefa] + " + " + this.instancia.getProc()[numMaq][tarefa]);
                custoAux += this.instancia.getPrep()[numMaq][tarefa][tarefa] + (this.instancia.getProc()[numMaq][tarefa]/instancia.getV()[this.speed[tarefa]]);
            }
            else{
                //System.out.println(this.instancia.getPrep()[numMaq][ultimaTarefa][tarefa] + " + " + this.instancia.getProc()[numMaq][tarefa]);
                custoAux += this.instancia.getPrep()[numMaq][ultimaTarefa][tarefa] + (this.instancia.getProc()[numMaq][tarefa]/instancia.getV()[this.speed[tarefa]]);
            }
            ultimaTarefa = tarefa;
        }
        return custoAux;
    }
    
    //Calcula o custo da máquina, usando uma máquina passada como parâmetro
    public double calcularCustoMaqList(ArrayList<Integer> maq, int numMaq, Instancia instancia){
    
        double custoAux = 0;
        int ultimaTarefa = 0;
        for(int j=0; j<maq.size(); j++){
            int tarefa = maq.get(j);
            if(j == 0){
                custoAux += instancia.getPrep()[numMaq][tarefa][tarefa] + (instancia.getProc()[numMaq][tarefa]/instancia.getV()[this.speed[tarefa]]);
            }
            else{
                custoAux += instancia.getPrep()[numMaq][ultimaTarefa][tarefa] + (instancia.getProc()[numMaq][tarefa]/instancia.getV()[this.speed[tarefa]]);
            }
            ultimaTarefa = tarefa;
        }
        return custoAux;
       
    }
    
    //Clona dus soluções
    public void clonarSolucoes(Solucao solRef){
        
        //Clona os Pesos
        this.peso_Cmax = solRef.getPeso_Cmax();
        this.peso_TEC  = solRef.getPeso_TEC();
        
        //Clona os z*
        this.z_Cmax = solRef.getZ_Cmax();
        this.z_Tec  = solRef.getZ_Tec();
        
        //Clona o Makespan
        this.setMakespan(solRef.calcularMakespan());
        
        //Clonando os custos das Máquina
        ArrayList<Double> aux = new ArrayList<Double>();
        for(int i=0; i< solRef.getCustoMaq().size(); i++){
            aux.add(solRef.getCustoMaq().get(i));
        }
        this.setCustoMaq(aux);
       
        //Clona as alocações
        for(int i=0; i<solRef.getSol().size();i++){
            ArrayList<Integer> auxArrayList = new ArrayList<Integer>();
            for(int j=0; j<solRef.getSol().get(i).size();j++){
                auxArrayList.add(solRef.getSol().get(i).get(j));
            }
            this.sol.remove(i);
            this.sol.add(i,auxArrayList);
        }
        
        //Novas características
        
        //Clona vetor de índices de velocidades
        int [] new_speed = new int[instancia.getNumJobs()];
        for(int i=0; i<instancia.getNumJobs(); i++){
            new_speed[i] = solRef.getSpeed()[i];
        }
        this.setSpeed(new_speed);
        
        //Clona TEC
        this.setTEC(solRef.calcularTEC());
        
    }


    public ArrayList<Double> getCustoMaq() {
        return custoMaq;
    }

    public void setCustoMaq(ArrayList<Double> custoMaq) {
        this.custoMaq = custoMaq;
    }
    
    @Override
    public int compareTo(Solucao o) {
        
        if(this.calcularMakespan() < o.calcularMakespan()){
              return -1;
        }
        if(this.calcularMakespan() > o.calcularMakespan()){
              return 1;
        }
        else{
            return 0;
        }
    }
    
    @Override
    public boolean equals(Object o) {  
        
        Solucao aux = (Solucao) o;
        if (this.makespan == aux.makespan)  
            return true;
        else
            return false;
    } 
            
}
